@@ -45,6 +45,8 @@ NAMESPACE = 'urn:ietf:params:xml:ns:caldav'
 # https://tools.ietf.org/html/rfc4791, section 4.2
 CALENDAR_RESOURCE_TYPE = '{%s}calendar' % NAMESPACE
 
+SUBSCRIPTION_RESOURCE_TYPE = '{http://calendarserver.org/ns/}subscribed'
+
 # TODO(jelmer): These resource types belong in scheduling.py
 SCHEDULE_INBOX_RESOURCE_TYPE = '{%s}schedule-inbox' % NAMESPACE
 SCHEDULE_OUTBOX_RESOURCE_TYPE = '{%s}schedule-outbox' % NAMESPACE
@@ -61,27 +63,27 @@ class Calendar(webdav.Collection):
     resource_types = (webdav.Collection.resource_types +
                       [CALENDAR_RESOURCE_TYPE])
 
-    def get_calendar_description(self):
+    def get_calendar_description(self) -> str:
         """Return the calendar description."""
         raise NotImplementedError(self.get_calendar_description)
 
-    def get_calendar_color(self):
+    def get_calendar_color(self) -> str:
         """Return the calendar color."""
         raise NotImplementedError(self.get_calendar_color)
 
-    def set_calendar_color(self, color):
+    def set_calendar_color(self, color: str) -> None:
         """Set the calendar color."""
         raise NotImplementedError(self.set_calendar_color)
 
-    def get_calendar_order(self):
+    def get_calendar_order(self) -> str:
         """Return the calendar order."""
         raise NotImplementedError(self.get_calendar_order)
 
-    def set_calendar_order(self, order):
+    def set_calendar_order(self, order: str) -> None:
         """Set the calendar order."""
         raise NotImplementedError(self.set_calendar_order)
 
-    def get_calendar_timezone(self):
+    def get_calendar_timezone(self) -> str:
         """Return calendar timezone.
 
         This should be an iCalendar object with exactly one
@@ -89,7 +91,7 @@ class Calendar(webdav.Collection):
         """
         raise NotImplementedError(self.get_calendar_timezone)
 
-    def set_calendar_timezone(self, content):
+    def set_calendar_timezone(self, content: str) -> None:
         """Set calendar timezone.
 
         This should be an iCalendar object with exactly one
@@ -97,14 +99,14 @@ class Calendar(webdav.Collection):
         """
         raise NotImplementedError(self.set_calendar_timezone)
 
-    def get_supported_calendar_components(self):
+    def get_supported_calendar_components(self) -> str:
         """Return set of supported calendar components in this calendar.
 
         :return: iterable over component names
         """
         raise NotImplementedError(self.get_supported_calendar_components)
 
-    def get_supported_calendar_data_types(self):
+    def get_supported_calendar_data_types(self) -> str:
         """Return supported calendar data types.
 
         :return: iterable over (content_type, version) tuples
@@ -163,6 +165,39 @@ class Calendar(webdav.Collection):
         raise NotImplementedError(self.calendar_query)
 
 
+class Subscription(object):
+
+    resource_types = (webdav.Collection.resource_types +
+                      [SUBSCRIPTION_RESOURCE_TYPE])
+
+    def get_source_url(self):
+        """Get the source URL for this calendar."""
+        raise NotImplementedError(self.get_source_url)
+
+    def set_source_url(self, url):
+        """Set the source URL for this calendar."""
+        raise NotImplementedError(self.set_source_url)
+
+    def get_calendar_description(self):
+        """Return the calendar description."""
+        raise NotImplementedError(self.get_calendar_description)
+
+    def get_calendar_color(self):
+        """Return the calendar color."""
+        raise NotImplementedError(self.get_calendar_color)
+
+    def set_calendar_color(self, color):
+        """Set the calendar color."""
+        raise NotImplementedError(self.set_calendar_color)
+
+    def get_supported_calendar_components(self):
+        """Return set of supported calendar components in this calendar.
+
+        :return: iterable over component names
+        """
+        raise NotImplementedError(self.get_supported_calendar_components)
+
+
 class CalendarHomeSet(object):
 
     def get_managed_attachments_server_url(self):
@@ -199,7 +234,7 @@ class CalendarHomeSetProperty(webdav.Property):
     in_allprops = False
     live = True
 
-    def get_value(self, base_href, resource, el, environ):
+    async def get_value(self, base_href, resource, el, environ):
         for href in resource.get_calendar_home_set():
             href = webdav.ensure_trailing_slash(href)
             el.append(webdav.create_href(href, base_href))
@@ -212,16 +247,38 @@ class CalendarDescriptionProperty(webdav.Property):
     """
 
     name = '{%s}calendar-description' % NAMESPACE
-    resource_type = CALENDAR_RESOURCE_TYPE
+    resource_type = (CALENDAR_RESOURCE_TYPE, SUBSCRIPTION_RESOURCE_TYPE)
 
-    def get_value(self, base_href, resource, el, environ):
+    async def get_value(self, base_href, resource, el, environ):
         el.text = resource.get_calendar_description()
 
     # TODO(jelmer): allow modification of this property
-    def set_value(self, href, resource, el):
+    async def set_value(self, href, resource, el):
         raise NotImplementedError
 
 
+def _extract_from_component(incomp, outcomp, requested):
+    for tag in requested:
+        if tag.tag == ('{%s}comp' % NAMESPACE):
+            for insub in incomp.subcomponents:
+                if insub.name == tag.get('name'):
+                    outsub = component_factory[insub.name]()
+                    outcomp.add_component(outsub)
+                    _extract_from_component(insub, outsub, tag)
+        elif tag.tag == ('{%s}prop' % NAMESPACE):
+            outcomp[tag.get('name')] = incomp[tag.get('name')]
+        elif tag.tag == ('{%s}allprop' % NAMESPACE):
+            for propname in incomp:
+                outcomp[propname] = incomp[propname]
+        elif tag.tag == ('{%s}allcomp' % NAMESPACE):
+            for insub in incomp.subcomponents:
+                outsub = component_factory[insub.name]()
+                outcomp.add_component(outsub)
+                _extract_from_component(insub, outsub, tag)
+        else:
+            raise AssertionError('invalid element %r' % tag)
+
+
 def extract_from_calendar(incal, outcal, requested):
     """Extract requested components/properties from calendar.
 
@@ -229,17 +286,11 @@ def extract_from_calendar(incal, outcal, requested):
     :param outcal: Calendar to write to
     :param requested: <calendar-data> element with requested
         components/properties
-    :return: A Calendar
     """
     for tag in requested:
         if tag.tag == ('{%s}comp' % NAMESPACE):
-            for insub in incal.subcomponents:
-                if insub.name == tag.get('name'):
-                    outsub = component_factory[insub.name]
-                    outcal.add_component(outsub)
-                    extract_from_calendar(insub, outsub, tag)
-        elif tag.tag == ('{%s}prop' % NAMESPACE):
-            outcal[tag.get('name')] = incal[tag.get('name')]
+            if incal.name == tag.get('name'):
+                _extract_from_component(incal, outcal, tag)
         elif tag.tag == ('{%s}expand' % NAMESPACE):
             # TODO(jelmer): https://github.com/jelmer/xandikos/issues/102
             raise NotImplementedError('expand is not yet implemented')
@@ -269,9 +320,9 @@ class CalendarDataProperty(davcommon.SubbedProperty):
     def supported_on(self, resource):
         return (resource.get_content_type() == 'text/calendar')
 
-    def get_value_ext(self, base_href, resource, el, environ, requested):
+    async def get_value_ext(self, base_href, resource, el, environ, requested):
         if len(requested) == 0:
-            serialized_cal = b''.join(resource.get_body())
+            serialized_cal = b''.join(await resource.get_body())
         else:
             c = ICalendar()
             calendar = calendar_from_resource(resource)
@@ -291,10 +342,10 @@ class CalendarOrderProperty(webdav.Property):
     name = '{http://apple.com/ns/ical/}calendar-order'
     resource_type = CALENDAR_RESOURCE_TYPE
 
-    def get_value(self, base_href, resource, el, environ):
+    async def get_value(self, base_href, resource, el, environ):
         el.text = resource.get_calendar_order()
 
-    def set_value(self, href, resource, el):
+    async def set_value(self, href, resource, el):
         resource.set_calendar_order(el.text)
 
 
@@ -376,7 +427,7 @@ def parse_time_range(el, cls):
     return cls(start, end)
 
 
-def parse_comp_filter(el, cls):
+def parse_comp_filter(el: ET.Element, cls):
     """Compile a comp-filter element into a Python function.
     """
     name = el.get('name')
@@ -405,7 +456,7 @@ def parse_comp_filter(el, cls):
     return comp_filter
 
 
-def parse_filter(filter_el, cls):
+def parse_filter(filter_el: ET.Element, cls):
     for subel in filter_el:
         if subel.tag == '{urn:ietf:params:xml:ns:caldav}comp-filter':
             parse_comp_filter(subel, cls.filter_subcomponent)
@@ -448,8 +499,9 @@ class CalendarQueryReporter(webdav.Reporter):
     data_property = CalendarDataProperty()
 
     @webdav.multistatus
-    def report(self, environ, body, resources_by_hrefs, properties, base_href,
-               base_resource, depth):
+    async def report(
+            self, environ, body, resources_by_hrefs, properties, base_href,
+            base_resource, depth):
         # TODO(jelmer): Verify that resource is a calendar
         requested = None
         filter_el = None
@@ -477,7 +529,7 @@ class CalendarQueryReporter(webdav.Reporter):
                 collection.calendar_query(filter_fn),
                 collection.subcollections())
 
-        for (href, resource) in webdav.traverse_resource(
+        async for (href, resource) in webdav.traverse_resource(
                 base_resource, base_href, depth,
                 members=members):
             # Ideally traverse_resource would only return the right things.
@@ -485,7 +537,8 @@ class CalendarQueryReporter(webdav.Reporter):
                 propstat = davcommon.get_properties_with_data(
                     self.data_property, href, resource, properties, environ,
                     requested)
-                yield webdav.Status(href, '200 OK', propstat=list(propstat))
+                yield webdav.Status(
+                    href, '200 OK', propstat=[s async for s in propstat])
 
 
 class CalendarColorProperty(webdav.Property):
@@ -495,12 +548,12 @@ class CalendarColorProperty(webdav.Property):
     """
 
     name = '{http://apple.com/ns/ical/}calendar-color'
-    resource_type = CALENDAR_RESOURCE_TYPE
+    resource_type = (CALENDAR_RESOURCE_TYPE, SUBSCRIPTION_RESOURCE_TYPE)
 
-    def get_value(self, href, resource, el, environ):
+    async def get_value(self, href, resource, el, environ):
         el.text = resource.get_calendar_color()
 
-    def set_value(self, href, resource, el):
+    async def set_value(self, href, resource, el):
         resource.set_calendar_color(el.text)
 
 
@@ -515,11 +568,12 @@ class SupportedCalendarComponentSetProperty(webdav.Property):
     name = '{%s}supported-calendar-component-set' % NAMESPACE
     resource_type = (CALENDAR_RESOURCE_TYPE,
                      SCHEDULE_INBOX_RESOURCE_TYPE,
-                     SCHEDULE_OUTBOX_RESOURCE_TYPE)
+                     SCHEDULE_OUTBOX_RESOURCE_TYPE,
+                     SUBSCRIPTION_RESOURCE_TYPE)
     in_allprops = False
     live = True
 
-    def get_value(self, href, resource, el, environ):
+    async def get_value(self, href, resource, el, environ):
         for component in resource.get_supported_calendar_components():
             subel = ET.SubElement(el, '{urn:ietf:params:xml:ns:caldav}comp')
             subel.set('name', component)
@@ -537,7 +591,7 @@ class SupportedCalendarDataProperty(webdav.Property):
                      SCHEDULE_OUTBOX_RESOURCE_TYPE)
     in_allprops = False
 
-    def get_value(self, href, resource, el, environ):
+    async def get_value(self, href, resource, el, environ):
         for (content_type, version) in (
                 resource.get_supported_calendar_data_types()):
             subel = ET.SubElement(
@@ -557,10 +611,10 @@ class CalendarTimezoneProperty(webdav.Property):
                      SCHEDULE_INBOX_RESOURCE_TYPE)
     in_allprops = False
 
-    def get_value(self, href, resource, el, environ):
+    async def get_value(self, href, resource, el, environ):
         el.text = resource.get_calendar_timezone()
 
-    def set_value(self, href, resource, el):
+    async def set_value(self, href, resource, el):
         if el is not None:
             resource.set_calendar_timezone(el.text)
         else:
@@ -580,7 +634,7 @@ class MinDateTimeProperty(webdav.Property):
     in_allprops = False
     live = True
 
-    def get_value(self, href, resource, el, environ):
+    async def get_value(self, href, resource, el, environ):
         el.text = resource.get_min_date_time()
 
 
@@ -597,7 +651,7 @@ class MaxDateTimeProperty(webdav.Property):
     in_allprops = False
     live = True
 
-    def get_value(self, href, resource, el, environ):
+    async def get_value(self, href, resource, el, environ):
         el.text = resource.get_max_date_time()
 
 
@@ -613,7 +667,7 @@ class MaxInstancesProperty(webdav.Property):
     in_allprops = False
     live = True
 
-    def get_value(self, href, resource, el, environ):
+    async def get_value(self, href, resource, el, environ):
         el.text = str(resource.get_max_instances())
 
 
@@ -630,7 +684,7 @@ class MaxAttendeesPerInstanceProperty(webdav.Property):
     in_allprops = False
     live = True
 
-    def get_value(self, href, resource, el, environ):
+    async def get_value(self, href, resource, el, environ):
         el.text = str(resource.get_max_attendees_per_instance())
 
 
@@ -647,7 +701,7 @@ class MaxResourceSizeProperty(webdav.Property):
     in_allprops = False
     live = True
 
-    def get_value(self, href, resource, el, environ):
+    async def get_value(self, href, resource, el, environ):
         el.text = str(resource.get_max_resource_size())
 
 
@@ -662,7 +716,7 @@ class MaxAttachmentsPerResourceProperty(webdav.Property):
     in_allprops = False
     live = True
 
-    def get_value(self, href, resource, el, environ):
+    async def get_value(self, href, resource, el, environ):
         el.text = str(resource.get_max_attachments_per_resource())
 
 
@@ -677,7 +731,7 @@ class MaxAttachmentSizeProperty(webdav.Property):
     in_allprops = False
     live = True
 
-    def get_value(self, href, resource, el, environ):
+    async def get_value(self, href, resource, el, environ):
         el.text = str(resource.get_max_attachment_size())
 
 
@@ -690,7 +744,7 @@ class ManagedAttachmentsServerURLProperty(webdav.Property):
     name = '{%s}managed-attachments-server-URL' % NAMESPACE
     in_allprops = False
 
-    def get_value(self, base_href, resource, el, environ):
+    async def get_value(self, base_href, resource, el, environ):
         # The RFC specifies that this property can be set on a calendar home
         # collection.
         # However, there is no matching resource type and we don't want to
@@ -704,6 +758,21 @@ class ManagedAttachmentsServerURLProperty(webdav.Property):
             el.append(webdav.create_href(href, base_href))
 
 
+class SourceProperty(webdav.Property):
+    """source property.
+    """
+    name = '{http://calendarserver.org/ns/}source'
+    resource_type = SUBSCRIPTION_RESOURCE_TYPE
+    in_allprops = True
+    live = False
+
+    async def get_value(self, base_href, resource, el, environ):
+        el.append(webdav.create_href(resource.get_source_url(), base_href))
+
+    async def set_value(self, href, resource, el):
+        raise NotImplementedError(self.set_value)
+
+
 class CalendarProxyReadForProperty(webdav.Property):
     """calendar-proxy-read-for property.
 
@@ -716,7 +785,7 @@ class CalendarProxyReadForProperty(webdav.Property):
     in_allprops = False
     live = True
 
-    def get_value(self, base_href, resource, el, environ):
+    async def get_value(self, base_href, resource, el, environ):
         for href in resource.get_calendar_proxy_read_for():
             el.append(webdav.create_href(href, base_href))
 
@@ -733,7 +802,7 @@ class CalendarProxyWriteForProperty(webdav.Property):
     in_allprops = False
     live = True
 
-    def get_value(self, base_href, resource, el, environ):
+    async def get_value(self, base_href, resource, el, environ):
         for href in resource.get_calendar_proxy_write_for():
             el.append(webdav.create_href(href, base_href))
 
@@ -748,7 +817,7 @@ class ScheduleCalendarTransparencyProperty(webdav.Property):
     live = False
     resource_type = CALENDAR_RESOURCE_TYPE
 
-    def get_value(self, base_href, resource, el, environ):
+    async def get_value(self, base_href, resource, el, environ):
         transp = resource.get_schedule_calendar_transparency()
         if transp == TRANSPARENCY_TRANSPARENT:
             ET.SubElement(el, '{%s}transparent' % NAMESPACE)
@@ -789,8 +858,8 @@ def extract_freebusy(comp, tzify):
     return ret
 
 
-def iter_freebusy(resources, start, end, tzify):
-    for (href, resource) in resources:
+async def iter_freebusy(resources, start, end, tzify):
+    async for (href, resource) in resources:
         c = calendar_from_resource(resource)
         if c is None:
             continue
@@ -813,8 +882,8 @@ class FreeBusyQueryReporter(webdav.Reporter):
     name = '{urn:ietf:params:xml:ns:caldav}free-busy-query'
     resource_type = CALENDAR_RESOURCE_TYPE
 
-    def report(self, environ, start_response, body, resources_by_hrefs,
-               properties, base_href, base_resource, depth):
+    async def report(self, environ, body, resources_by_hrefs,
+                     properties, base_href, base_resource, depth):
         requested = None
         for el in body:
             if el.tag == '{urn:ietf:params:xml:ns:caldav}time-range':
@@ -835,47 +904,43 @@ class FreeBusyQueryReporter(webdav.Reporter):
         fb['DTSTAMP'] = vDDDTypes(tzify(datetime.datetime.now()))
         fb['DTSTART'] = vDDDTypes(start)
         fb['DTEND'] = vDDDTypes(end)
-        fb['FREEBUSY'] = list(iter_freebusy(
+        fb['FREEBUSY'] = [item async for item in iter_freebusy(
             webdav.traverse_resource(base_resource, base_href, depth),
-            start, end, tzify))
+            start, end, tzify)]
         ret.add_component(fb)
-        start_response('200 OK', [])
-        return [ret.to_ical()]
+        return webdav.Response(status='200 OK', body=[ret.to_ical()])
 
 
 class MkcalendarMethod(webdav.Method):
 
-    def handle(self, environ, start_response, app):
-        try:
-            content_type = environ['CONTENT_TYPE']
-        except KeyError:
-            base_content_type = None
-        else:
-            base_content_type, params = webdav.parse_type(content_type)
+    async def handle(self, request, environ, app):
+        content_type = request.content_type
+        base_content_type, params = webdav.parse_type(content_type)
         if base_content_type not in (
-            'text/xml', 'application/xml', None, 'text/plain'
+            'text/xml', 'application/xml', None, 'text/plain',
+            'application/octet-stream',
         ):
             raise webdav.UnsupportedMediaType(content_type)
-        href, path, resource = app._get_resource_from_environ(environ)
+        href, path, resource = app._get_resource_from_environ(request, environ)
         if resource is not None:
             return webdav._send_simple_dav_error(
-                environ, start_response,
+                request,
                 '403 Forbidden',
                 error=ET.Element('{DAV:}resource-must-be-null'),
                 description=('Something already exists at %r' % path))
         try:
             resource = app.backend.create_collection(path)
         except FileNotFoundError:
-            start_response('409 Conflict', [])
-            return []
+            return webdav.Response(status='409 Conflict')
         el = ET.Element('{DAV:}resourcetype')
-        app.properties['{DAV:}resourcetype'].get_value(
+        await app.properties['{DAV:}resourcetype'].get_value(
             href, resource, el, environ)
         ET.SubElement(el, '{urn:ietf:params:xml:ns:caldav}calendar')
         app.properties['{DAV:}resourcetype'].set_value(href, resource, el)
         if base_content_type in ('text/xml', 'application/xml'):
-            et = webdav._readXmlBody(
-                environ, '{urn:ietf:params:xml:ns:caldav}mkcalendar')
+            et = await webdav._readXmlBody(
+                request, '{urn:ietf:params:xml:ns:caldav}mkcalendar',
+                strict=app.strict)
             propstat = []
             for el in et:
                 if el.tag != '{DAV:}set':
@@ -888,7 +953,6 @@ class MkcalendarMethod(webdav.Method):
             for propstat_el in webdav.propstat_as_xml(propstat):
                 ret.append(propstat_el)
             return webdav._send_xml_response(
-                start_response, '201 Created', ret, webdav.DEFAULT_ENCODING)
+                '201 Created', ret, webdav.DEFAULT_ENCODING)
         else:
-            start_response('201 Created', [])
-            return []
+            return webdav.Response(status='201 Created')
