@@ -31,6 +31,14 @@ import jinja2
 import logging
 import os
 import posixpath
+from typing import (
+    List,
+    Tuple,
+    Iterable,
+    Iterator,
+    Optional,
+)
+
 import shutil
 import urllib.parse
 
@@ -42,15 +50,20 @@ from xandikos.icalendar import (
     CalendarFilter,
 )
 from xandikos.store import (
+    Store,
+    File,
     DuplicateUidError,
     InvalidFileContents,
     NoSuchItem,
     NotStoreError,
+    LockedError,
+    OutOfSpaceError,
     STORE_TYPE_ADDRESSBOOK,
     STORE_TYPE_CALENDAR,
     STORE_TYPE_PRINCIPAL,
     STORE_TYPE_SCHEDULE_INBOX,
     STORE_TYPE_SCHEDULE_OUTBOX,
+    STORE_TYPE_SUBSCRIPTION,
     STORE_TYPE_OTHER,
 )
 from xandikos.store.git import (
@@ -69,27 +82,34 @@ CALENDAR_HOME_SET = ['calendars']
 ADDRESSBOOK_HOME_SET = ['contacts']
 
 TEMPLATES_DIR = os.path.join(os.path.dirname(__file__), 'templates')
-jinja_env = jinja2.Environment(loader=jinja2.FileSystemLoader(TEMPLATES_DIR))
+jinja_env = jinja2.Environment(
+    loader=jinja2.FileSystemLoader(TEMPLATES_DIR),
+    enable_async=True)
 
 
-def render_jinja_page(name, accepted_content_languages, **kwargs):
+async def render_jinja_page(
+        name: str, accepted_content_languages: List[str],
+        **kwargs) -> Tuple[
+            Iterable[bytes], int, Optional[str], str, List[str]]:
     """Render a HTML page from jinja template.
 
     :param name: Name of the page
     :param accepted_content_languages: List of accepted content languages
-    :return: TUple of (body, content_length, etag, content_type, languages)
+    :return: Tuple of (body, content_length, etag, content_type, languages)
     """
     # TODO(jelmer): Support rendering other languages
     encoding = 'utf-8'
     template = jinja_env.get_template(name)
-    body = template.render(
+    body = await template.render_async(
         version=xandikos_version,
-        urljoin=urllib.parse.urljoin, **kwargs).encode(encoding)
-    return ([body], len(body), None, 'text/html; encoding=%s' % encoding,
+        urljoin=urllib.parse.urljoin, **kwargs)
+    body_encoded = body.encode(encoding)
+    return ([body_encoded], len(body_encoded),
+            None, 'text/html; encoding=%s' % encoding,
             ['en-UK'])
 
 
-def create_strong_etag(etag):
+def create_strong_etag(etag: str) -> str:
     """Create strong etags.
 
     :param etag: basic etag
@@ -98,7 +118,7 @@ def create_strong_etag(etag):
     return '"' + etag + '"'
 
 
-def extract_strong_etag(etag):
+def extract_strong_etag(etag: Optional[str]) -> Optional[str]:
     """Extract a strong etag from a string."""
     if etag is None:
         return etag
@@ -108,27 +128,28 @@ def extract_strong_etag(etag):
 class ObjectResource(webdav.Resource):
     """Object resource."""
 
-    def __init__(self, store, name, content_type, etag, file=None):
+    def __init__(self, store: Store, name: str, content_type: str, etag: str,
+                 file: Optional[File] = None):
         self.store = store
         self.name = name
         self.etag = etag
         self.content_type = content_type
         self._file = file
 
-    def __repr__(self):
+    def __repr__(self) -> str:
         return "%s(%r, %r, %r, %r)" % (
             type(self).__name__, self.store, self.name, self.etag,
             self.get_content_type()
         )
 
     @property
-    def file(self):
+    def file(self) -> File:
         if self._file is None:
             self._file = self.store.get_file(self.name, self.content_type,
                                              self.etag)
         return self._file
 
-    def get_body(self):
+    async def get_body(self) -> Iterable[bytes]:
         return self.file.content
 
     def set_body(self, data, replace_etag=None):
@@ -147,16 +168,16 @@ class ObjectResource(webdav.Resource):
                 'UID already in use.')
         return create_strong_etag(etag)
 
-    def get_content_language(self):
+    def get_content_language(self) -> str:
         raise KeyError
 
-    def get_content_type(self):
+    def get_content_type(self) -> str:
         return self.content_type
 
-    def get_content_length(self):
-        return sum(map(len, self.get_body()))
+    async def get_content_length(self) -> int:
+        return sum(map(len, await self.get_body()))
 
-    def get_etag(self):
+    async def get_etag(self) -> str:
         return create_strong_etag(self.etag)
 
     def get_supported_locks(self):
@@ -225,34 +246,40 @@ class StoreBasedCollection(object):
             self.store.set_type(STORE_TYPE_SCHEDULE_OUTBOX)
         elif resource_types == {webdav.COLLECTION_RESOURCE_TYPE}:
             self.store.set_type(STORE_TYPE_OTHER)
+        elif resource_types == {webdav.COLLECTION_RESOURCE_TYPE,
+                                caldav.SUBSCRIPTION_RESOURCE_TYPE}:
+            self.store.set_type(STORE_TYPE_SUBSCRIPTION)
         else:
             raise NotImplementedError(self.set_resource_types)
 
-    def _get_resource(self, name, content_type, etag, file=None):
+    def _get_resource(self, name: str,
+                      content_type: str,
+                      etag: str,
+                      file: Optional[File] = None) -> webdav.Resource:
         return ObjectResource(self.store, name, content_type, etag, file=file)
 
-    def _get_subcollection(self, name):
+    def _get_subcollection(self, name: str) -> webdav.Collection:
         return self.backend.get_resource(posixpath.join(self.relpath, name))
 
-    def get_displayname(self):
+    def get_displayname(self) -> str:
         displayname = self.store.get_displayname()
         if displayname is None:
             return os.path.basename(self.store.repo.path)
         return displayname
 
-    def set_displayname(self, displayname):
+    def set_displayname(self, displayname: str) -> None:
         self.store.set_displayname(displayname)
 
-    def get_sync_token(self):
+    def get_sync_token(self) -> str:
         return self.store.get_ctag()
 
-    def get_ctag(self):
+    def get_ctag(self) -> str:
         return self.store.get_ctag()
 
-    def get_etag(self):
+    async def get_etag(self) -> str:
         return create_strong_etag(self.store.get_ctag())
 
-    def members(self):
+    def members(self) -> Iterator[Tuple[str, webdav.Resource]]:
         for (name, content_type, etag) in self.store.iter_with_etag():
             resource = self._get_resource(name, content_type, etag)
             yield (name, resource)
@@ -281,7 +308,8 @@ class StoreBasedCollection(object):
             # self.get_subcollection(name).destroy()
             shutil.rmtree(os.path.join(self.store.path, name))
 
-    def create_member(self, name, contents, content_type):
+    def create_member(self, name: str, contents: Iterable[bytes],
+                      content_type: str) -> Tuple[str, str]:
         try:
             (name, etag) = self.store.import_one(name, content_type, contents)
         except InvalidFileContents as e:
@@ -293,9 +321,17 @@ class StoreBasedCollection(object):
             raise webdav.PreconditionFailure(
                 '{%s}no-uid-conflict' % caldav.NAMESPACE,
                 'UID already in use.')
+        except OutOfSpaceError:
+            raise webdav.InsufficientStorage()
+        except LockedError:
+            raise webdav.ResourceLocked()
         return (name, create_strong_etag(etag))
 
-    def iter_differences_since(self, old_token, new_token):
+    def iter_differences_since(
+            self, old_token: str, new_token: str) -> Iterator[Tuple[
+                str, Optional[webdav.Resource], Optional[webdav.Resource]]]:
+        old_resource: Optional[webdav.Resource]
+        new_resource: Optional[webdav.Resource]
         for (name, content_type,
              old_etag, new_etag) in self.store.iter_changes(
                  old_token, new_token):
@@ -341,26 +377,26 @@ class StoreBasedCollection(object):
     def get_content_language(self):
         raise KeyError
 
-    def get_content_length(self):
+    async def get_content_length(self):
         raise KeyError
 
-    def destroy(self):
+    def destroy(self) -> None:
         # RFC2518, section 8.6.2 says this should recursively delete.
         self.store.destroy()
 
-    def get_body(self):
+    async def get_body(self):
         raise NotImplementedError(self.get_body)
 
-    def render(self, self_url, accepted_content_types,
-               accepted_content_languages):
+    async def render(self, self_url, accepted_content_types,
+                     accepted_content_languages):
         content_types = webdav.pick_content_types(
             accepted_content_types, ['text/html'])
         assert content_types == ['text/html']
-        return render_jinja_page(
+        return await render_jinja_page(
             'collection.html', accepted_content_languages, collection=self,
             self_url=self_url)
 
-    def get_is_executable(self):
+    def get_is_executable(self) -> bool:
         return False
 
     def get_quota_used_bytes(self):
@@ -392,6 +428,35 @@ class ScheduleOutbox(StoreBasedCollection, scheduling.ScheduleOutbox):
     """A schedling outbox collection."""
 
 
+class SubscriptionCollection(StoreBasedCollection, caldav.Subscription):
+
+    def get_source_url(self):
+        source_url = self.store.get_source_url()
+        if source_url is None:
+            raise KeyError
+        return source_url
+
+    def set_source_url(self, url):
+        self.store.set_source_url(url)
+
+    def get_calendar_description(self):
+        return self.store.get_description()
+
+    def get_calendar_color(self):
+        color = self.store.get_color()
+        if not color:
+            raise KeyError
+        if color and color[0] != '#':
+            color = '#' + color
+        return color
+
+    def set_calendar_color(self, color):
+        self.store.set_color(color)
+
+    def get_supported_calendar_components(self):
+        return ["VEVENT", "VTODO", "VJOURNAL", "VFREEBUSY"]
+
+
 class CalendarCollection(StoreBasedCollection, caldav.Calendar):
 
     def get_calendar_description(self):
@@ -524,7 +589,7 @@ class CollectionSetResource(webdav.Collection):
     def get_sync_token(self):
         raise KeyError
 
-    def get_etag(self):
+    async def get_etag(self):
         raise KeyError
 
     def get_ctag(self):
@@ -567,7 +632,7 @@ class CollectionSetResource(webdav.Collection):
     def get_content_language(self):
         raise KeyError
 
-    def get_content_length(self):
+    async def get_content_length(self):
         raise KeyError
 
     def get_last_modified(self):
@@ -583,12 +648,12 @@ class CollectionSetResource(webdav.Collection):
         # RFC2518, section 8.6.2 says this should recursively delete.
         shutil.rmtree(p)
 
-    def render(self, self_url, accepted_content_types,
-               accepted_content_languages):
+    async def render(self, self_url, accepted_content_types,
+                     accepted_content_languages):
         content_types = webdav.pick_content_types(
             accepted_content_types, ['text/html'])
         assert content_types == ['text/html']
-        return render_jinja_page(
+        return await render_jinja_page(
             'root.html', accepted_content_languages, self_url=self_url)
 
     def get_is_executable(self):
@@ -602,11 +667,15 @@ class CollectionSetResource(webdav.Collection):
         # TODO(jelmer): Ask the store?
         raise KeyError
 
+    def get_creationdate(self):
+        # TODO(jelmer): Find creation date using store function
+        raise KeyError
+
 
 class RootPage(webdav.Resource):
     """A non-DAV resource."""
 
-    resource_types = []
+    resource_types: List[str] = []
 
     def __init__(self, backend):
         self.backend = backend
@@ -621,10 +690,10 @@ class RootPage(webdav.Resource):
             principals=self.backend.find_principals(),
             self_url=self_url)
 
-    def get_body(self):
+    async def get_body(self):
         raise KeyError
 
-    def get_content_length(self):
+    async def get_content_length(self):
         raise KeyError
 
     def get_content_type(self):
@@ -636,9 +705,9 @@ class RootPage(webdav.Resource):
     def get_active_locks(self):
         return []
 
-    def get_etag(self):
+    async def get_etag(self):
         h = hashlib.md5()
-        for c in self.get_body():
+        for c in await self.get_body():
             h.update(c)
         return h.hexdigest()
 
@@ -649,11 +718,11 @@ class RootPage(webdav.Resource):
         return ['en-UK']
 
     def get_member(self, name):
-        return self.backend.get_resource(name)
+        return self.backend.get_resource('/' + name)
 
     def delete_member(self, name, etag=None):
         # This doesn't have any non-collection members.
-        self.get_member(name).destroy()
+        self.get_member('/' + name).destroy()
 
     def get_is_executable(self):
         return False
@@ -754,12 +823,12 @@ class PrincipalBare(CollectionSetResource, Principal):
                 pass
         return p
 
-    def render(self, self_url, accepted_content_types,
-               accepted_content_languages):
+    async def render(self, self_url, accepted_content_types,
+                     accepted_content_languages):
         content_types = webdav.pick_content_types(
             accepted_content_types, ['text/html'])
         assert content_types == ['text/html']
-        return render_jinja_page(
+        return await render_jinja_page(
             'principal.html', accepted_content_languages, principal=self,
             self_url=self_url)
 
@@ -825,6 +894,8 @@ class XandikosBackend(webdav.Backend):
 
     def get_resource(self, relpath):
         relpath = posixpath.normpath(relpath)
+        if not relpath.startswith('/'):
+            raise ValueError('relpath %r should start with /')
         if relpath == '/':
             return RootPage(self)
         p = self._map_to_file_path(relpath)
@@ -844,6 +915,7 @@ class XandikosBackend(webdav.Backend):
                     STORE_TYPE_PRINCIPAL: PrincipalCollection,
                     STORE_TYPE_SCHEDULE_INBOX: ScheduleInbox,
                     STORE_TYPE_SCHEDULE_OUTBOX: ScheduleOutbox,
+                    STORE_TYPE_SUBSCRIPTION: SubscriptionCollection,
                     STORE_TYPE_OTHER: Collection
                 }[store.get_type()](self, relpath, store)
         else:
@@ -864,8 +936,8 @@ class XandikosApp(webdav.WebDAVApp):
     """A wsgi App that provides a Xandikos web server.
     """
 
-    def __init__(self, backend, current_user_principal):
-        super(XandikosApp, self).__init__(backend)
+    def __init__(self, backend, current_user_principal, strict=True):
+        super(XandikosApp, self).__init__(backend, strict=strict)
 
         def get_current_user_principal(env):
             try:
@@ -882,6 +954,7 @@ class XandikosApp(webdav.WebDAVApp):
             webdav.GetContentTypeProperty(),
             webdav.GetContentLengthProperty(),
             webdav.GetContentLanguageProperty(),
+            caldav.SourceProperty(),
             caldav.CalendarHomeSetProperty(),
             carddav.AddressbookHomeSetProperty(),
             caldav.CalendarDescriptionProperty(),
@@ -947,24 +1020,6 @@ class XandikosApp(webdav.WebDAVApp):
         ])
 
 
-class WellknownRedirector(object):
-    """Redirect paths under .well-known/ to the appropriate paths."""
-
-    def __init__(self, inner_app, dav_root):
-        self._inner_app = inner_app
-        self._dav_root = dav_root
-
-    def __call__(self, environ, start_response):
-        # See https://tools.ietf.org/html/rfc6764
-        path = posixpath.normpath(
-            environ['SCRIPT_NAME'] + environ['PATH_INFO'])
-        if path in WELLKNOWN_DAV_PATHS:
-            start_response('302 Found', [
-                ('Location', self._dav_root)])
-            return []
-        return self._inner_app(environ, start_response)
-
-
 def create_principal_defaults(backend, principal):
     """Create default calendar and addressbook for a principal.
 
@@ -1002,6 +1057,42 @@ def create_principal_defaults(backend, principal):
         logging.info('Create inbox in %s.', resource.store.path)
 
 
+class RedirectDavHandler(object):
+
+    def __init__(self, dav_root: str):
+        self._dav_root = dav_root
+
+    async def __call__(self, request):
+        from aiohttp import web
+        return web.HTTPFound(self._dav_root)
+
+
+MDNS_NAME = 'Xandikos CalDAV/CardDAV service'
+
+
+def avahi_register(port: int, path: str):
+    import avahi
+    import dbus
+    bus = dbus.SystemBus()
+    server = dbus.Interface(
+        bus.get_object(avahi.DBUS_NAME, avahi.DBUS_PATH_SERVER),
+        avahi.DBUS_INTERFACE_SERVER)
+    group = dbus.Interface(
+        bus.get_object(avahi.DBUS_NAME, server.EntryGroupNew()),
+        avahi.DBUS_INTERFACE_ENTRY_GROUP)
+
+    for service in ['_carddav._tcp', '_caldav._tcp']:
+        try:
+            group.AddService(
+                avahi.IF_UNSPEC, avahi.PROTO_INET, 0, MDNS_NAME, service,
+                '', '', port,
+                avahi.string_array_to_txt_array(['path=%s' % path]))
+        except dbus.DBusException as e:
+            logging.error('Error registering %s: %s', service, e)
+
+    group.Commit()
+
+
 def main(argv):
     import argparse
     import sys
@@ -1017,8 +1108,9 @@ def main(argv):
 
     access_group = parser.add_argument_group(title="Access Options")
     access_group.add_argument(
-        "-l", "--listen_address", dest="listen_address", default="localhost",
-        help="Binding IP address. [%(default)s]")
+        "-l", "--listen-address", dest="listen_address", default="localhost",
+        help=("Bind to this address. "
+              "Pass in path for unix domain socket. [%(default)s]"))
     access_group.add_argument(
         "-p", "--port", dest="port", type=int, default=8080,
         help="Port to listen on. [%(default)s]")
@@ -1043,6 +1135,13 @@ def main(argv):
     parser.add_argument(
         "--dump-dav-xml", action="store_true", dest="dump_dav_xml",
         help="Print DAV XML request/responses.")
+    parser.add_argument(
+        '--avahi', action='store_true',
+        help='Announce services with avahi.')
+    parser.add_argument(
+        '--no-strict', action='store_false', dest='strict',
+        help="Enable workarounds for buggy CalDAV/CardDAV client "
+        "implementations.", default=True)
     options = parser.parse_args(argv[1:])
 
     if options.directory is None:
@@ -1076,26 +1175,63 @@ def main(argv):
             'Run xandikos with --autocreate?',
             options.current_user_principal)
 
-    app = XandikosApp(
+    main_app = XandikosApp(
         backend,
-        current_user_principal=options.current_user_principal)
+        current_user_principal=options.current_user_principal,
+        strict=options.strict)
 
-    from wsgiref.simple_server import make_server
-    app = WellknownRedirector(app, options.route_prefix)
-    server = make_server(options.listen_address, options.port, app)
-    logging.info('Listening on %s:%s', options.listen_address,
-                 options.port)
+    async def xandikos_handler(request):
+        return await main_app.aiohttp_handler(request, options.route_prefix)
 
-    import signal
-    # Set SIGINT to default handler; this appears to be necessary
-    # when running under coverage.
-    signal.signal(signal.SIGINT, signal.SIG_DFL)
-    try:
-        server.serve_forever()
-    except KeyboardInterrupt:
-        pass
+    if '/' in options.listen_address:
+        socket_path = options.listen_address
+        listen_address = None
+        logging.info('Listening on unix domain socket %s', socket_path)
     else:
-        server.shutdown()
+        listen_address = options.listen_address
+        socket_path = None
+        logging.info('Listening on %s:%s', listen_address, options.port)
+
+    from aiohttp import web
+
+    app = web.Application()
+    try:
+        import prometheus_client  # noqa: F401
+    except ModuleNotFoundError:
+        logging.warning(
+            'Prometheus client not found; /metrics will not be available.')
+    else:
+        from .metrics import setup_metrics
+        setup_metrics(app)
+
+    for path in WELLKNOWN_DAV_PATHS:
+        app.router.add_route(
+            "*", path,
+            RedirectDavHandler(options.route_prefix).__call__)
+
+    if options.route_prefix.strip('/'):
+        xandikos_app = web.Application()
+        xandikos_app.router.add_route("*", "/{path_info:.*}", xandikos_handler)
+
+        async def redirect_to_subprefix(request):
+            return web.HTTPFound(options.route_prefix)
+        app.router.add_route("*", "/", redirect_to_subprefix)
+        app.add_subapp(options.route_prefix, xandikos_app)
+    else:
+        app.router.add_route("*", "/{path_info:.*}", xandikos_handler)
+
+    if options.avahi:
+        try:
+            import avahi  # noqa: F401
+            import dbus   # noqa: F401
+        except ImportError:
+            logging.error(
+                'Please install python-avahi and python-dbus for '
+                'avahi support.')
+        else:
+            avahi_register(options.port, options.route_prefix)
+
+    web.run_app(app, port=options.port, host=listen_address, path=socket_path)
 
 
 if __name__ == '__main__':
