@@ -33,7 +33,10 @@ import logging
 import os
 import posixpath
 from typing import (
+    Callable,
     Dict,
+    Iterable,
+    AsyncIterable,
     List,
     Optional,
     Union,
@@ -197,6 +200,14 @@ class PreconditionFailure(Exception):
         self.description = description
 
 
+class InsufficientStorage(Exception):
+    """Insufficient storage."""
+
+
+class ResourceLocked(Exception):
+    """Resource locked."""
+
+
 def etag_matches(condition, actual_etag):
     """Check if an etag matches an If-Matches condition.
 
@@ -959,7 +970,9 @@ class Principal(Resource):
         raise NotImplementedError(self.get_schedule_outbox_url)
 
 
-async def get_property_from_name(href, resource, properties, name, environ):
+async def get_property_from_name(
+        href: str, resource: Resource,
+        properties, name: str, environ):
     """Get a single property on a resource.
 
     :param href: Resource href
@@ -973,8 +986,10 @@ async def get_property_from_name(href, resource, properties, name, environ):
         href, resource, properties, environ, ET.Element(name))
 
 
-async def get_property_from_element(href, resource, properties, environ,
-                                    requested):
+async def get_property_from_element(
+        href: str, resource: Resource,
+        properties: Dict[str, Property],
+        environ, requested: ET.Element) -> PropStatus:
     """Get a single property on a resource.
 
     :param href: Resource href
@@ -998,19 +1013,27 @@ async def get_property_from_element(href, resource, properties, environ,
             if not prop.supported_on(resource):
                 raise KeyError
             try:
-                get_value_ext = prop.get_value_ext
+                get_value_ext = prop.get_value_ext  # type: ignore
             except AttributeError:
                 await prop.get_value(href, resource, ret, environ)
             else:
                 await get_value_ext(href, resource, ret, environ, requested)
         except KeyError:
             statuscode = '404 Not Found'
+        except NotImplementedError:
+            logging.exception(
+                'Not implemented while getting %s for %r',
+                requested.tag, resource)
+            statuscode = '501 Not Implemented'
         else:
             statuscode = '200 OK'
     return PropStatus(statuscode, responsedescription, ret)
 
 
-async def get_properties(href, resource, properties, environ, requested):
+async def get_properties(
+        href: str, resource: Resource,
+        properties: Dict[str, Property], environ,
+        requested: ET.Element) -> AsyncIterable[PropStatus]:
     """Get a set of properties.
 
     :param href: Resource Href
@@ -1025,7 +1048,10 @@ async def get_properties(href, resource, properties, environ, requested):
             href, resource, properties, environ, propreq)
 
 
-async def get_property_names(href, resource, properties, environ, requested):
+async def get_property_names(
+        href: str, resource: Resource,
+        properties: Dict[str, Property], environ,
+        requested: ET.Element) -> AsyncIterable[PropStatus]:
     """Get a set of property names.
 
     :param href: Resource Href
@@ -1040,7 +1066,9 @@ async def get_property_names(href, resource, properties, environ, requested):
             yield PropStatus('200 OK', None, ET.Element(name))
 
 
-async def get_all_properties(href, resource, properties, environ):
+async def get_all_properties(
+        href: str, resource: Resource,
+        properties: Dict[str, Property], environ) -> AsyncIterable[PropStatus]:
     """Get all properties.
 
     :param href: Resource Href
@@ -1057,7 +1085,7 @@ async def get_all_properties(href, resource, properties, environ):
             yield ps
 
 
-def ensure_trailing_slash(href):
+def ensure_trailing_slash(href: str) -> str:
     """Ensure that a href has a trailing slash.
 
     Useful for collection hrefs, e.g. when used with urljoin.
@@ -1070,7 +1098,11 @@ def ensure_trailing_slash(href):
     return href + '/'
 
 
-async def traverse_resource(base_resource, base_href, depth, members=None):
+async def traverse_resource(
+        base_resource: Resource, base_href: str, depth: str,
+        members: Optional[
+            Callable[[Collection], Iterable[Tuple[str, Resource]]]] = None
+        ) -> AsyncIterable[Tuple[str, Resource]]:
     """Traverse a resource.
 
     :param base_resource: Resource to traverse from
@@ -1081,8 +1113,10 @@ async def traverse_resource(base_resource, base_href, depth, members=None):
     :return: Iterator over (URL, Resource) tuples
     """
     if members is None:
-        def members(c):
+        def members_fn(c):
             return c.members()
+    else:
+        members_fn = members
     todo = collections.deque([(base_href, base_resource, depth)])
     while todo:
         (href, resource, depth) = todo.popleft()
@@ -1102,7 +1136,7 @@ async def traverse_resource(base_resource, base_href, depth, members=None):
         else:
             raise AssertionError("invalid depth %r" % depth)
         if COLLECTION_RESOURCE_TYPE in resource.resource_types:
-            for (child_name, child_resource) in members(resource):
+            for (child_name, child_resource) in members_fn(resource):
                 child_href = urllib.parse.urljoin(href, child_name)
                 todo.append((child_href, child_resource, nextdepth))
 
@@ -1127,14 +1161,15 @@ class Reporter(object):
                        for rs in self.resource_type)
         return self.resource_type in resource.resource_types
 
-    async def report(self,
-                     environ: Dict[str, str],
-                     request_body: ET.Element,
-                     resources_by_hrefs,
-                     properties: Dict[str, Property],
-                     href: str,
-                     resource: Resource,
-                     depth: str):
+    async def report(
+            self, environ: Dict[str, str],
+            request_body: ET.Element,
+            resources_by_hrefs: Callable[
+                [Iterable[str]], Iterable[Tuple[str, Resource]]],
+            properties: Dict[str, Property],
+            href: str,
+            resource: Resource,
+            depth: str) -> Status:
         """Send a report.
 
         :param environ: wsgi environ
@@ -1177,8 +1212,12 @@ class ExpandPropertyReporter(Reporter):
 
     name = '{DAV:}expand-property'
 
-    async def _populate(self, prop_list, resources_by_hrefs, properties, href,
-                        resource, environ):
+    async def _populate(
+            self, prop_list: ET.Element,
+            resources_by_hrefs: Callable[
+                [Iterable[str]], List[Tuple[str, Resource]]],
+            properties: Dict[str, Property], href: str,
+            resource: Resource, environ) -> AsyncIterable[Status]:
         """Expand properties for a resource.
 
         :param prop_list: DAV:property elements to retrieve and expand
@@ -1192,30 +1231,37 @@ class ExpandPropertyReporter(Reporter):
         ret = []
         for prop in prop_list:
             prop_name = prop.get('name')
+            if prop_name is None:
+                logging.warning('Tag %s without name attribute', prop.tag)
+                continue
             # FIXME: Resolve prop_name on resource
             propstat = await get_property_from_name(
                 href, resource, properties, prop_name, environ)
             new_prop = ET.Element(propstat.prop.tag)
-            child_hrefs = [
+            child_hrefs = filter(None, [
                 read_href_element(prop_child)
                 for prop_child in propstat.prop
-                if prop_child.tag == '{DAV:}href']
+                if prop_child.tag == '{DAV:}href'])
             child_resources = resources_by_hrefs(child_hrefs)
             for prop_child in propstat.prop:
                 if prop_child.tag != '{DAV:}href':
                     new_prop.append(prop_child)
                 else:
                     child_href = read_href_element(prop_child)
-                    child_resource = child_resources[child_href]
+                    if child_href is None:
+                        logging.warning(
+                            'Tag %s without valid href', prop_child.tag)
+                        continue
+                    child_resource = dict(child_resources).get(child_href)
                     if child_resource is None:
                         # FIXME: What to do if the referenced href is invalid?
                         # For now, let's just keep the unresolved href around
                         new_prop.append(prop_child)
                     else:
-                        response = self._populate(
-                            prop, properties, child_href, child_resource,
-                            environ)
-                        new_prop.append(response.aselement())
+                        async for response in self._populate(
+                                prop, resources_by_hrefs, properties,
+                                child_href, child_resource, environ):
+                            new_prop.append(response.aselement())
             propstat = PropStatus(propstat.statuscode,
                                   propstat.responsedescription, prop=new_prop)
             ret.append(propstat)
@@ -1305,7 +1351,7 @@ class Backend(object):
         """
         raise NotImplementedError(self.create_collection)
 
-    def get_resoure(self, relpath):
+    def get_resource(self, relpath):
         raise NotImplementedError(self.get_resource)
 
 
@@ -1435,10 +1481,11 @@ async def _readBody(request):
         return [await request.content.read(request_body_size)]
 
 
-async def _readXmlBody(request, expected_tag=None):
+async def _readXmlBody(
+        request, expected_tag: Optional[str] = None, strict: bool = True):
     content_type = request.content_type
     base_content_type, params = parse_type(content_type)
-    if base_content_type not in ('text/xml', 'application/xml'):
+    if strict and base_content_type not in ('text/xml', 'application/xml'):
         raise UnsupportedMediaType(content_type)
     body = b''.join(await _readBody(request))
     if os.environ.get('XANDIKOS_DUMP_DAV_XML'):
@@ -1504,6 +1551,10 @@ class PostMethod(Method):
                 request, '412 Precondition Failed',
                 error=ET.Element(e.precondition),
                 description=e.description)
+        except InsufficientStorage:
+            return Response(status=507, reason='Insufficient Storage')
+        except ResourceLocked:
+            return Response(status=423, reason='Resource Locked')
         href = (
             environ['SCRIPT_NAME'] +
             urllib.parse.urljoin(ensure_trailing_slash(path), name)
@@ -1557,6 +1608,10 @@ class PutMethod(Method):
                 request, '412 Precondition Failed',
                 error=ET.Element(e.precondition),
                 description=e.description)
+        except InsufficientStorage:
+            return Response(status=507, reason='Insufficient Storage')
+        except ResourceLocked:
+            return Response(status=423, reason='Resource Locked')
         return Response(
             status=201, reason='Created', headers=[
                 ('ETag', new_etag)])
@@ -1571,7 +1626,7 @@ class ReportMethod(Method):
         if r is None:
             return _send_not_found(request)
         depth = request.headers.get("Depth", "0")
-        et = await _readXmlBody(request, None)
+        et = await _readXmlBody(request, None, strict=app.strict)
         try:
             reporter = app.reporters[et.tag]
         except KeyError:
@@ -1607,7 +1662,8 @@ class PropfindMethod(Method):
         if not request.can_read_body:
             requested = None
         else:
-            et = await _readXmlBody(request, '{DAV:}propfind')
+            et = await _readXmlBody(
+                request, '{DAV:}propfind', strict=app.strict)
             try:
                 [requested] = et
             except ValueError:
@@ -1643,7 +1699,8 @@ class ProppatchMethod(Method):
         if resource is None:
             yield Status(request.url, '404 Not Found')
             return
-        et = await _readXmlBody(request, '{DAV:}propertyupdate')
+        et = await _readXmlBody(
+            request, '{DAV:}propertyupdate', strict=app.strict)
         propstat = []
         for el in et:
             if el.tag not in ('{DAV:}set', '{DAV:}remove'):
@@ -1674,7 +1731,7 @@ class MkcolMethod(Method):
             return Response(status=409, reason='Conflict')
         if base_content_type in ('text/xml', 'application/xml'):
             # Extended MKCOL (RFC5689)
-            et = await _readXmlBody(request, '{DAV:}mkcol')
+            et = await _readXmlBody(request, '{DAV:}mkcol', strict=app.strict)
             propstat = []
             for el in et:
                 if el.tag != '{DAV:}set':
@@ -1819,11 +1876,12 @@ class WebDAVApp(object):
     (returning None for nonexistant objects).
     """
 
-    def __init__(self, backend):
+    def __init__(self, backend, strict=True):
         self.backend = backend
         self.properties = {}
         self.reporters = {}
         self.methods = {}
+        self.strict = strict
         self.register_methods([
             DeleteMethod(),
             PostMethod(),
